from pathlib import Path
from json import load, dump
from typing import Dict

from jsonschema import Draft7Validator
from jsonschema.validators import validator_for, validate


def validate_schema_definition(schema: Dict):
    """ TODO: validate the schema itself """


def schema_from_json(path: Path):
    with open(path, 'r') as _input:
        schema = load(_input)
    validate_schema_definition(schema)
    return schema


def is_ref(schema: dict):
    return '$ref' in schema


class Schema:
    """ Combines schema-related functionality """

    def __init__(self, schema: Dict, validator_cls=None):
        self.schema: Dict = schema

        if validator_cls is None:
            validator_cls = validator_for(schema)
        self.validator_cls = validator_cls
        self.check_schema()

    def check_schema(self):
        self.validator_cls.check_schema(self.schema)

    def is_valid_data(self, data):
        try:
            validate(data, self.schema)
            return True
        except Exception:
            return False

    def validator(self, schema: Dict = None):
        if schema is None:
            schema = self.schema
        return self.validator_cls(schema)

    def resolve_schema(self, schema: Dict) -> Dict:
        """ resolve reference schema """
        if is_ref(schema):
            ident = schema['$ref']
            _first, definition, key = ident.split('/')
            return self.schema[definition][key]

        return schema

    def definitions(self):
        return self.schema.get('definitions', {})

    @staticmethod
    def _enum_defaults(schema):
        try:
            return schema["enum"][0]
        except IndexError:
            return None

    def _object_defaults(self, schema):
        return {k: self.compute_defaults(s) for k, s in schema["properties"].items()}

    def _any_of_defaults(self, schema, definitions: Dict = None):
        out = [self.compute_defaults(s) for s in schema["anyOf"]]
        return out

    def _array_defaults(self, schema):
        items_schema = schema['items']
        if isinstance(items_schema, dict):
            return []

        return [self.compute_defaults(s) for s in schema["items"]]

    def compute_defaults(self, schema):
        schema = self.resolve_schema(schema)

        if "default" in schema:
            return schema["default"]

        if 'anyOf' in schema:
            return self._any_of_defaults(schema)

        if "enum" in schema:
            return self._enum_defaults(schema)

        schema_type = schema.get("type")

        if schema_type == "object":
            return self._object_defaults(schema)

        elif schema_type == "array":
            return self._array_defaults(schema)

        return None



from typing import List, Dict

from PyQt5.QtWidgets import QWidget, QVBoxLayout, QGroupBox, QLabel

from pyqtschema.widgets.base import SchemaWidgetMixin


class FormWidget(QWidget):

    def __init__(self, widget: SchemaWidgetMixin, *args, **kwargs):
        super().__init__(*args, **kwargs)
        layout = QVBoxLayout()
        self.setLayout(layout)

        self.error_widget = QGroupBox(parent=self)
        self.error_widget.setTitle("Errors")
        self.error_layout = QVBoxLayout()
        self.error_widget.setLayout(self.error_layout)
        self.error_widget.hide()

        layout.addWidget(self.error_widget)
        layout.addWidget(widget)

        self.widget = widget

    def display_errors(self, errors: List[Exception]):
        self.error_widget.show()

        layout = self.error_widget.layout()
        while True:
            item = layout.takeAt(0)
            if not item:
                break
            item.widget().deleteLater()

        for err in errors:
            widget = QLabel(f"<b>.{'.'.join(err.path)}</b> {err.message}")
            layout.addWidget(widget)

    def clear_errors(self):
        self.error_widget.hide()

    @property
    def state(self) -> Dict:
        return self.widget.state

    @state.setter
    def state(self, state: dict):
        self.widget.state = state



from PyQt5.QtWidgets import QCheckBox

from pyqtschema.widgets.base import SchemaWidgetMixin, state_property


class CheckboxSchemaWidget(SchemaWidgetMixin, QCheckBox):

    @state_property
    def state(self) -> bool:
        return self.isChecked()

    @state.setter
    def state(self, checked: bool):
        self.setChecked(checked)

    def configure(self):
        self.stateChanged.connect(lambda _: self.on_changed.emit(self.state))



from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QPushButton, QColorDialog

from pyqtschema.widgets.base import SchemaWidgetMixin, state_property


class QColorButton(QPushButton):
    """Color picker widget QPushButton subclass.

    Implementation derived from https://martinfitzpatrick.name/article/qcolorbutton-a-color-selector-tool-for-pyqt/
    """

    colorChanged = pyqtSignal()

    def __init__(self, *args, **kwargs):
        super(QColorButton, self).__init__(*args, **kwargs)

        self._color = None
        self.pressed.connect(self.onColorPicker)

    def color(self):
        return self._color

    def setColor(self, color):
        if color != self._color:
            self._color = color
            self.colorChanged.emit()

        if self._color:
            self.setStyleSheet("background-color: %s;" % self._color)
        else:
            self.setStyleSheet("")

    def onColorPicker(self):
        dlg = QColorDialog(self)
        if self._color:
            dlg.setCurrentColor(QColor(self._color))

        if dlg.exec_():
            self.setColor(dlg.currentColor().name())

    def mousePressEvent(self, event):
        if event.button() == Qt.RightButton:
            self.setColor(None)

        return super(QColorButton, self).mousePressEvent(event)


class ColorSchemaWidget(SchemaWidgetMixin, QColorButton):
    """Widget representation of a string with the 'color' format keyword."""

    def configure(self):
        self.colorChanged.connect(lambda: self.on_changed.emit(self.state))

    @state_property
    def state(self) -> str:
        return self.color()

    @state.setter
    def state(self, data: str):
        self.setColor(data)



from enum import Enum

from PyQt5.QtWidgets import QComboBox

from pyqtschema.widgets.base import SchemaWidgetMixin, state_property


class EnumSchemaWidget(SchemaWidgetMixin, QComboBox):

    @state_property
    def state(self):
        return self.itemData(self.currentIndex())

    @state.setter
    def state(self, value):
        val = value
        if isinstance(val, Enum):
            val = val.value
        index = self.findData(val)
        if index == -1:
            raise ValueError(val)
        self.setCurrentIndex(index)

    def configure(self):
        options = self.schema["enum"]
        for idx, opt in enumerate(options):
            vis, val = str(opt), opt
            self.addItem(vis)
            self.setItemData(idx, val)

        self.currentIndexChanged.connect(lambda _: self.on_changed.emit(self.state))

    def _index_changed(self, index: int):
        self.on_changed.emit(self.state)



from typing import Iterator

from PyQt5.QtWidgets import QLayoutItem, QWidget, QLayout


def is_concrete_schema(schema: dict) -> bool:
    return "type" in schema


def iter_layout_items(layout) -> Iterator[QLayoutItem]:
    return (layout.itemAt(i) for i in range(layout.count()))


def iter_layout_widgets(layout: QLayout) -> Iterator[QWidget]:
    return (i.widget() for i in iter_layout_items(layout))



from collections import OrderedDict
from functools import partial
from typing import Dict, List

from PyQt5.QtWidgets import QWidget, QVBoxLayout, QComboBox, QGroupBox

from ..schema import Schema
from .base import state_property, SchemaWidgetMixin
from .utils import iter_layout_widgets

CHECKER_KEY = 'checker'


def _any_of_schema(schema: dict, widget_builder: 'WidgetBuilder') -> List[Schema]:
    """ """
    out = []
    for _any in schema['anyOf']:
        _schema = {'type': 'object',
                   'properties': {
                       CHECKER_KEY: widget_builder.schema.resolve_schema(_any)
                   },
                   'definitions': widget_builder.schema.definitions(),
                   "schema": {
                       'additionalProperties': False,
                   },
                   'additionalProperties': False,
                   }
        out.append(Schema(_schema))

    return out


class AnyOfSchemaWidget(SchemaWidgetMixin, QGroupBox):

    def __init__(self, schema: dict, ui_schema: dict, widget_builder: 'WidgetBuilder', *args, **kwargs):
        self.select_combo: QComboBox = None
        self.items_group: QGroupBox = None

        super().__init__(schema, ui_schema, widget_builder, *args, **kwargs)

        self.widget_labels: List = []
        self.widgets: List = []
        self.populate_from_schema(schema, ui_schema, widget_builder)

        self.any_of_schemas: List[Schema] = _any_of_schema(schema, widget_builder)

    def configure(self):
        if 'title' in self.schema:
            self.setTitle(self.schema['title'])

        self.select_combo = QComboBox()
        self.select_combo.currentIndexChanged.connect(self.select_schema)

        lay = QVBoxLayout()
        self.items_group = QGroupBox()
        self.items_group.setLayout(lay)

        layout = QVBoxLayout()
        layout.addWidget(self.select_combo)
        layout.addWidget(self.items_group)
        self.setLayout(layout)

    def select_schema(self, idx: int):
        for iii, widget in enumerate(iter_layout_widgets(self.items_group.layout())):
            widget.setVisible(iii == idx)
        self.on_changed.emit(self.state)

    def populate_from_schema(self, schema: dict, ui_schema: dict, widget_builder: 'WidgetBuilder'
                             ) -> Dict[str, QWidget]:
        self.setFlat(False)

        _combo_items = []
        for sub_schema in schema.get('anyOf', []):
            try:  # TODO: Use BBuilderClass to unify the determination
                name = sub_schema['$ref']
                c_name = name.split('/')[-1]
            except KeyError:
                c_name = name = sub_schema['type']

            _combo_items.append(c_name)
            sub_ui_schema = ui_schema.get(name, {})

            widget = widget_builder.create_widget(sub_schema, sub_ui_schema, parent=self)  # TODO onchanged
            self.widget_labels.append(c_name)
            self.widgets.append(widget)
            widget.setVisible(False)

            widget.on_changed.connect(partial(self.widget_on_changed, name))
            if 'title' in schema:
                self.setTitle(schema['title'])

            if 'description' in schema:
                self.setToolTip(schema['description'])

            self.items_group.layout().addWidget(widget)

        self.select_combo.addItems(_combo_items)
        return self.widgets

    def widget_on_changed(self, name: str, value):
        # self.state[name] = value
        self.on_changed.emit(self.state)

    @state_property
    def state(self) -> dict:
        _idx = self.select_combo.currentIndex()
        return self.widgets[_idx].state

    @state.setter
    def state(self, state: dict):
        # assumption:
        #   state is dict => set the state to the current selection
        #   state is a list => set state to all widgets
        # Background:
        #   the function may be called after initializing the form including all defaults for any possible item
        _widgets = self.widgets
        if isinstance(state, (list, tuple)):
            for idx, _dat in enumerate(state):
                if idx >= len(_widgets):
                    break
                _widgets[idx].state = _dat

        else:
            _done = False
            for idx, _schema in enumerate(self.any_of_schemas):
                if isinstance(state, dict):
                    # TODO: improve the workaround (which does not work for nested structures)
                    # verify key manually, since additionalProperties is not working (due to a bug? in jsonschema)
                    stop = False
                    for key in state.keys():
                        stop = key not in _schema.schema['properties'].get(CHECKER_KEY, {}).get('properties', {})
                        if stop:
                            break
                    if stop:
                        continue

                if _schema.is_valid_data({CHECKER_KEY: state}):
                    _done = True
                    _widgets[idx].state = state
                    self.select_combo.setCurrentIndex(idx)
                    break

            if not _done:
                _idx = self.select_combo.currentIndex()
                _widgets[_idx].state = state

    def show_title(self) -> bool:
        return False



from functools import partial
from typing import List, Tuple, Optional

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QWidget, QHBoxLayout, QVBoxLayout, QPushButton, QStyle, QGroupBox

from pyqtschema.widgets.base import SchemaWidgetMixin, state_property
from pyqtschema.widgets.utils import iter_layout_widgets, is_concrete_schema


class ArrayControlsWidget(QWidget):
    on_delete = pyqtSignal()
    on_move_up = pyqtSignal()
    on_move_down = pyqtSignal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        style = self.style()

        self.up_button = QPushButton()
        self.up_button.setIcon(style.standardIcon(QStyle.SP_ArrowUp))
        self.up_button.clicked.connect(lambda _: self.on_move_up.emit())

        self.delete_button = QPushButton()
        self.delete_button.setIcon(style.standardIcon(QStyle.SP_DialogCancelButton))
        self.delete_button.clicked.connect(lambda _: self.on_delete.emit())

        self.down_button = QPushButton()
        self.down_button.setIcon(style.standardIcon(QStyle.SP_ArrowDown))
        self.down_button.clicked.connect(lambda _: self.on_move_down.emit())

        group_layout = QHBoxLayout()
        self.setLayout(group_layout)
        group_layout.addWidget(self.up_button)
        group_layout.addWidget(self.down_button)
        group_layout.addWidget(self.delete_button)
        group_layout.setSpacing(0)
        group_layout.addStretch(0)


class ArrayRowWidget(QWidget):

    def __init__(self, widget: QWidget, controls: ArrayControlsWidget, *args, **kwargs):
        super().__init__(*args, **kwargs)

        layout = QHBoxLayout()
        layout.addWidget(widget)
        layout.addWidget(controls)
        layout.setStretch(0, 5)
        self.setLayout(layout)

        self.widget = widget
        self.controls = controls


class ArraySchemaWidget(SchemaWidgetMixin, QGroupBox):

    def show_title(self) -> bool:
        return False

    @property
    def rows(self) -> List[ArrayRowWidget]:
        return [*iter_layout_widgets(self.array_layout)]

    @state_property
    def state(self) -> list:
        return [r.widget.state for r in self.rows]

    @state.setter
    def state(self, state: list):
        for row in self.rows:
            self._remove_item(row)

        for item in state:
            self._add_item(item)

        self.on_changed.emit(self.state)

    def handle_error(self, path: Tuple[str], err: Exception):
        index, *tail = path
        self.rows[index].widget.handle_error(tail, err)

    def configure(self):
        layout = QVBoxLayout()
        style = self.style()

        if 'title' in self.schema:
            self.setTitle(self.schema['title'])

        self.add_button = QPushButton()
        self.add_button.setIcon(style.standardIcon(QStyle.SP_DialogOkButton))
        self.add_button.clicked.connect(lambda _: self.add_item())

        self.array_layout = QVBoxLayout()
        # self.array_layout.setContentsMargins(0, 0, 0, 0)
        array_widget = QWidget(self)
        array_widget.setLayout(self.array_layout)

        self.on_changed.connect(self._on_updated)

        layout.addWidget(self.add_button)
        layout.addWidget(array_widget)
        self.setLayout(layout)

    def _on_updated(self, state):
        # Update add button
        disabled = self.next_item_schema is None
        self.add_button.setEnabled(not disabled)

        previous_row = None
        for i, row in enumerate(self.rows):
            if previous_row:
                can_exchange_previous = previous_row.widget.schema == row.widget.schema
                row.controls.up_button.setEnabled(can_exchange_previous)
                previous_row.controls.down_button.setEnabled(can_exchange_previous)
            else:
                row.controls.up_button.setEnabled(False)
            row.controls.delete_button.setEnabled(not self.is_fixed_schema(i))
            previous_row = row

        if previous_row:
            previous_row.controls.down_button.setEnabled(False)

    def is_fixed_schema(self, index: int) -> bool:
        schema = self.schema['items']
        if isinstance(schema, dict):
            return False

        return index < len(schema)

    @property
    def next_item_schema(self) -> Optional[dict]:
        item_schema = self.schema['items']

        if isinstance(item_schema, dict):
            return item_schema

        index = len(self.rows)

        try:
            item_schema = item_schema[index]
        except IndexError:
            item_schema = self.schema.get("additionalItems", {})
            if isinstance(item_schema, bool):
                return None

        if not is_concrete_schema(item_schema):
            return None

        return item_schema

    def add_item(self, item_state=None):
        self._add_item(item_state)
        self.on_changed.emit(self.state)

    def remove_item(self, row: ArrayRowWidget):
        self._remove_item(row)
        self.on_changed.emit(self.state)

    def move_item_up(self, row: ArrayRowWidget):
        index = self.rows.index(row)
        self.array_layout.insertWidget(max(0, index - 1), row)
        self.on_changed.emit(self.state)

    def move_item_down(self, row: ArrayRowWidget):
        index = self.rows.index(row)
        self.array_layout.insertWidget(min(len(self.rows) - 1, index + 1), row)
        self.on_changed.emit(self.state)

    def _add_item(self, item_state=None):
        item_schema = self.next_item_schema

        # Create widget
        item_ui_schema = self.ui_schema.get("items", {})
        widget = self.widget_builder.create_widget(item_schema, item_ui_schema, item_state, parent=self)
        controls = ArrayControlsWidget()

        # Create row
        row = ArrayRowWidget(widget, controls)
        row.layout().setContentsMargins(0, 0, 0, 0)

        self.array_layout.addWidget(row)

        # Setup callbacks
        widget.on_changed.connect(partial(self.widget_on_changed, row))
        controls.on_delete.connect(partial(self.remove_item, row))
        controls.on_move_up.connect(partial(self.move_item_up, row))
        controls.on_move_down.connect(partial(self.move_item_down, row))

        return row

    def _remove_item(self, row: ArrayRowWidget):
        self.array_layout.removeWidget(row)
        row.deleteLater()

    def widget_on_changed(self, row: ArrayRowWidget, value):
        self.state[self.rows.index(row)] = value
        self.on_changed.emit(self.state)



from PyQt5.QtWidgets import QLineEdit, QTextEdit

from .base import SchemaWidgetMixin, state_property


class TextSchemaWidget(SchemaWidgetMixin, QLineEdit):
    """ """

    def configure(self):
        self.textChanged.connect(self.on_changed.emit)

    @state_property
    def state(self) -> str:
        return str(self.text())

    @state.setter
    def state(self, state: str):
        self.setText(state)


class TextAreaSchemaWidget(SchemaWidgetMixin, QTextEdit):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @state_property
    def state(self) -> str:
        return str(self.toPlainText())

    @state.setter
    def state(self, state: str):
        self.setPlainText(state)

    def configure(self):
        self.textChanged.connect(lambda: self.on_changed.emit(self.state))


class PasswordWidget(TextSchemaWidget):

    def configure(self):
        super().configure()

        self.setEchoMode(self.Password)
